// © 2022 Daily, Co. All Rights Reserved

import 'package:freezed_annotation/freezed_annotation.dart';

import 'update.dart';

part 'subscriptions.freezed.dart';
part 'subscriptions.g.dart';

/// The identity of a subscription profile.
///
/// Subscription profiles can be used to conveniently update multiple subscription settings for groups of participants.
@Freezed(toJson: false, fromJson: false)
class SubscriptionProfile with _$SubscriptionProfile {
  const factory SubscriptionProfile(String name) = _SubscriptionProfile;

  const SubscriptionProfile._();

  factory SubscriptionProfile.fromJson(String name) => SubscriptionProfile(name);

  static const SubscriptionProfile base = SubscriptionProfile('base');
  static const SubscriptionProfile activeRemote = SubscriptionProfile('active-remote');

  String toJson() => name;
}

/// Updates the subscription profile that is set for a remote participant.
@Freezed(toJson: false)
class SubscriptionProfileUpdate with _$SubscriptionProfileUpdate implements Update<SubscriptionProfile> {
  const SubscriptionProfileUpdate._();

  /// Sets the remote participant's profile.
  const factory SubscriptionProfileUpdate.set({required SubscriptionProfile profile}) = _SubscriptionProfileUpdate;

  /// Restores the remote participant's profile to the default ([SubscriptionProfile.base]).
  static const Update<SubscriptionProfileUpdate> fromDefaults = UpdateFromDefaults.instance;

  @override
  String toJson() => profile.toJson();
}

/// Subscription settings for a remote participant or profile.
@Freezed(copyWith: false)
class SubscriptionSettings with _$SubscriptionSettings {
  const factory SubscriptionSettings({
    required SubscriptionProfile profile,
    required MediaSubscriptionSettings media,
  }) = _SubscriptionSettings;

  factory SubscriptionSettings.fromJson(Map<String, dynamic> json) => _$SubscriptionSettingsFromJson(json);
}

/// Updates subscription settings for a remote participant or profile.
@Freezed(toJson: true)
class SubscriptionSettingsUpdate with _$SubscriptionSettingsUpdate implements Update<SubscriptionSettingsUpdate> {
  /// Updates subscription settings for a remote participant or profile with specific values.
  const factory SubscriptionSettingsUpdate.set({
    Update<SubscriptionProfile>? profile,
    Update<MediaSubscriptionSettingsUpdate>? media,
  }) = _SubscriptionSettingsUpdate;

  /// Restores subscription settings for a remote participant or profile to their default values.
  static const Update<SubscriptionSettingsUpdate> fromDefaults = UpdateFromDefaults.instance;
}

/// Media subscription settings for a remote participant or profile.
@Freezed(copyWith: false)
class MediaSubscriptionSettings with _$MediaSubscriptionSettings {
  const factory MediaSubscriptionSettings({
    required CameraSubscriptionSettings camera,
    required MicrophoneSubscriptionSettings microphone,
  }) = _MediaSubscriptionSettings;

  factory MediaSubscriptionSettings.fromJson(Map<String, dynamic> json) => _$MediaSubscriptionSettingsFromJson(json);
}

/// Updates media subscription settings for a remote participant or profile.
@Freezed(toJson: true)
class MediaSubscriptionSettingsUpdate
    with _$MediaSubscriptionSettingsUpdate
    implements Update<MediaSubscriptionSettingsUpdate> {
  /// Updates media subscription settings for a remote participant or profile with specific values.
  const factory MediaSubscriptionSettingsUpdate.set({
    Update<CameraSubscriptionSettingsUpdate>? camera,
    Update<MicrophoneSubscriptionSettingsUpdate>? microphone,
  }) = _MediaSubscriptionSettingsUpdate;

  /// Restores media subscription settings for a remote participant or profile to their default values.
  static const Update<MediaSubscriptionSettingsUpdate> fromDefaults = UpdateFromDefaults.instance;
}

/// Microphone subscription settings for a remote participant or profile.
@Freezed(copyWith: false)
class MicrophoneSubscriptionSettings with _$MicrophoneSubscriptionSettings {
  const factory MicrophoneSubscriptionSettings({
    required SubscriptionState subscriptionState,
  }) = _MicrophoneSubscriptionSettingsValue;

  factory MicrophoneSubscriptionSettings.fromJson(Map<String, dynamic> json) =>
      _$MicrophoneSubscriptionSettingsFromJson(json);
}

/// Updates microphone subscription settings for a remote participant or profile.
@Freezed(toJson: true)
class MicrophoneSubscriptionSettingsUpdate
    with _$MicrophoneSubscriptionSettingsUpdate
    implements Update<MicrophoneSubscriptionSettingsUpdate> {
  /// Updates microphone subscription settings for a remote participant or profile with specific values.
  const factory MicrophoneSubscriptionSettingsUpdate.set({
    SubscriptionStateUpdate? subscriptionState,
  }) = _MicrophoneSubscriptionSettingsUpdateValue;

  /// Restores microphone subscription settings for a remote participant or profile to their default values.
  static const Update<MicrophoneSubscriptionSettingsUpdate> fromDefaults = UpdateFromDefaults.instance;
}

/// Camera subscription settings for a remote participant or profile.
@Freezed(copyWith: false)
class CameraSubscriptionSettings with _$CameraSubscriptionSettings {
  const factory CameraSubscriptionSettings({
    required SubscriptionState subscriptionState,
    required VideoReceiveSettings receiveSettings,
  }) = _CameraSubscriptionSettingsValue;

  factory CameraSubscriptionSettings.fromJson(Map<String, dynamic> json) => _$CameraSubscriptionSettingsFromJson(json);
}

/// Updates camera subscription settings for a remote participant or profile.
@Freezed(toJson: true)
class CameraSubscriptionSettingsUpdate
    with _$CameraSubscriptionSettingsUpdate
    implements Update<CameraSubscriptionSettingsUpdate> {
  /// Updates camera subscription settings for a remote participant or profile with specific values.
  const factory CameraSubscriptionSettingsUpdate.set({
    SubscriptionStateUpdate? subscriptionState,
    Update<VideoReceiveSettingsUpdate>? receiveSettings,
  }) = _CameraSubscriptionSettingsUpdateValue;

  /// Restores camera subscription settings for a remote participant or profile to their default values.
  static const Update<CameraSubscriptionSettingsUpdate> fromDefaults = UpdateFromDefaults.instance;
}

/// Video receiving subscription settings for a remote participant or profile.
@Freezed(copyWith: false)
class VideoReceiveSettings with _$VideoReceiveSettings {
  const factory VideoReceiveSettings({
    required VideoReceiveSettingsMaxQuality maxQuality,
  }) = _VideoReceiveSettingsValue;

  factory VideoReceiveSettings.fromJson(Map<String, dynamic> json) => _$VideoReceiveSettingsFromJson(json);
}

/// Updates video receiving subscription settings for a remote participant or profile.
@Freezed(toJson: true)
class VideoReceiveSettingsUpdate with _$VideoReceiveSettingsUpdate implements Update<VideoReceiveSettingsUpdate> {
  const factory VideoReceiveSettingsUpdate.set({
    /// Updates video receiving subscription settings for a remote participant or profile with specific values.
    VideoReceiveSettingsMaxQualityUpdate? maxQuality,
  }) = _VideoReceiveSettingsUpdateValue;

  /// Restores video receiving subscription settings for a remote participant or profile to their default values.
  static const Update<VideoReceiveSettingsUpdate> fromDefaults = UpdateFromDefaults.instance;
}

enum VideoReceiveSettingsMaxQuality { low, medium, high }

enum VideoReceiveSettingsMaxQualityUpdate { fromDefaults, low, medium, high }

enum SubscriptionState { subscribed, unsubscribed }

enum SubscriptionStateUpdate { fromDefaults, subscribed, unsubscribed }
